/* $XFree86: xc/lib/GL/mesa/src/drv/r200/r200_context.c,v 1.3 2003/05/06 23:52:08 daenzer Exp $ */
/*
Copyright (C) The Weather Channel, Inc.  2002.  All Rights Reserved.

The Weather Channel (TM) funded Tungsten Graphics to develop the
initial release of the Radeon 8500 driver under the XFree86 license.
This notice must be preserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE COPYRIGHT OWNER(S) AND/OR ITS SUPPLIERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

**************************************************************************/

/*
 * Authors:
 *   Keith Whitwell <keith@tungstengraphics.com>
 */

#include "glheader.h"
#include "api_arrayelt.h"
#include "context.h"
#include "simple_list.h"
#include "imports.h"
#include "matrix.h"
#include "extensions.h"
#include "state.h"

#include "swrast/swrast.h"
#include "swrast_setup/swrast_setup.h"
#include "array_cache/acache.h"

#include "tnl/tnl.h"
#include "tnl/t_pipeline.h"

#include "drivers/common/driverfuncs.h"

#include "r200_context.h"
#include "radeon_ioctl.h"
#include "r200_state.h"
#include "radeon_span.h"
#include "r200_pixel.h"
#include "r200_tex.h"
#include "r200_swtcl.h"
#include "r200_tcl.h"
#include "r200_vtxfmt.h"
#include "r200_maos.h"

#include "vblank.h"
#include "utils.h"
#include "xmlpool.h"		/* for symbolic values of enum-type options */
//

/* Extension strings exported by the R200 driver.
 */
static const char *const card_extensions[] = {
	"GL_ARB_multisample",
	"GL_ARB_multitexture",
	"GL_ARB_texture_border_clamp",
	"GL_ARB_texture_compression",
	"GL_ARB_texture_env_add",
	"GL_ARB_texture_env_combine",
	"GL_ARB_texture_env_dot3",
	"GL_ARB_texture_mirrored_repeat",
	"GL_ARB_vertex_buffer_object",
	"GL_EXT_blend_minmax",
	"GL_EXT_blend_subtract",
	"GL_EXT_secondary_color",
	"GL_EXT_stencil_wrap",
	"GL_EXT_texture_edge_clamp",
	"GL_EXT_texture_env_combine",
	"GL_EXT_texture_env_dot3",
	"GL_EXT_texture_filter_anisotropic",
	"GL_EXT_texture_lod_bias",
	"GL_EXT_texture_mirror_clamp",
	"GL_EXT_texture_rectangle",
	"GL_ATI_texture_env_combine3",
	"GL_ATI_texture_mirror_once",
	"GL_MESA_pack_invert",
	"GL_MESA_ycbcr_texture",
	"GL_NV_blend_square",
	"GL_SGIS_generate_mipmap",
	NULL
};

extern const struct tnl_pipeline_stage _r200_render_stage;
extern const struct tnl_pipeline_stage _r200_tcl_stage;

static const struct tnl_pipeline_stage *r200_pipeline[] = {

	/* Try and go straight to t&l
	 */
	&_r200_tcl_stage,

	/* Catch any t&l fallbacks
	 */
	&_tnl_vertex_transform_stage,
	&_tnl_normal_transform_stage,
	&_tnl_lighting_stage,
	&_tnl_fog_coordinate_stage,
	&_tnl_texgen_stage,
	&_tnl_texture_transform_stage,
	&_tnl_vertex_program_stage,

	/* Try again to go to tcl?
	 *     - no good for asymmetric-twoside (do with multipass)
	 *     - no good for asymmetric-unfilled (do with multipass)
	 *     - good for material
	 *     - good for texgen
	 *     - need to manipulate a bit of state
	 *
	 * - worth it/not worth it?
	 */

	/* Else do them here.
	 */
	/*    &_r200_render_stage,  *//* FIXME: bugs with ut2003 */
	&_tnl_render_stage,	/* FALLBACK:  */
	0,
};


/* Create the device specific rendering context.
 */
GLboolean r200CreateContext(const __GLcontextModes * glVisual,
			    __DRIcontextPrivate * driContextPriv,
			    void *sharedContextPrivate)
{
	__DRIscreenPrivate *sPriv = driContextPriv->driScreenPriv;
	radeonScreenPtr screen = (radeonScreenPtr) (sPriv->private);
	struct dd_function_table functions;
	r200ContextPtr rmesa;
	GLcontext *ctx;
	int i;
	int tcl_mode;

	assert(glVisual);
	assert(driContextPriv);
	assert(screen);

	/* Allocate the R200 context */
	rmesa = (r200ContextPtr) CALLOC(sizeof(*rmesa));
	if (!rmesa)
		return GL_FALSE;

	/* Parse configuration files.
	 * Do this here so that initialMaxAnisotropy is set before we create
	 * the default textures.
	 */
	driParseConfigFiles(&rmesa->radeon.optionCache, &screen->optionCache,
			    screen->driScreen->myNum, "r200");
	rmesa->initialMaxAnisotropy = driQueryOptionf(&rmesa->radeon.optionCache,
						      "def_max_anisotropy");

	/* Init default driver functions then plug in our R200-specific functions
	 * (the texture functions are especially important)
	 */
	_mesa_init_driver_functions(&functions);
	r200InitIoctlFuncs(&functions);
	r200InitStateFuncs(&functions);
	r200InitTextureFuncs(&functions);

	if (!radeonInitContext(&rmesa->radeon, &functions,
			       glVisual, driContextPriv, sharedContextPrivate)) {
		FREE(rmesa);
		return GL_FALSE;
	}

	/* Init r200 context data */
	rmesa->dma.buf0_address = rmesa->radeon.radeonScreen->buffers->list[0].address;

	(void)memset(rmesa->texture_heaps, 0, sizeof(rmesa->texture_heaps));
	make_empty_list(&rmesa->swapped);

	rmesa->nr_heaps = 1 /* screen->numTexHeaps */ ;
	assert(rmesa->nr_heaps < R200_NR_TEX_HEAPS);
	for (i = 0; i < rmesa->nr_heaps; i++) {
		rmesa->texture_heaps[i] = driCreateTextureHeap(i, rmesa,
							       screen->
							       texSize[i], 12,
							       RADEON_NR_TEX_REGIONS,
							       (drmTextureRegionPtr)
							       rmesa->radeon.sarea->
							       tex_list[i],
							       &rmesa->radeon.sarea->
							       tex_age[i],
							       &rmesa->swapped,
							       sizeof
							       (r200TexObj),
							       (destroy_texture_object_t
								*)
							       r200DestroyTexObj);
	}
	rmesa->texture_depth = driQueryOptioni(&rmesa->radeon.optionCache,
					       "texture_depth");
	if (rmesa->texture_depth == DRI_CONF_TEXTURE_DEPTH_FB)
		rmesa->texture_depth = (screen->cpp == 4) ?
		    DRI_CONF_TEXTURE_DEPTH_32 : DRI_CONF_TEXTURE_DEPTH_16;

	rmesa->swtcl.RenderIndex = ~0;
	rmesa->hw.all_dirty = 1;

	/* Set the maximum texture size small enough that we can guarentee that
	 * all texture units can bind a maximal texture and have them both in
	 * texturable memory at once.
	 */

	ctx = rmesa->radeon.glCtx;
	ctx->Const.MaxTextureUnits = driQueryOptioni(&rmesa->radeon.optionCache,
						     "texture_units");
	ctx->Const.MaxTextureImageUnits = ctx->Const.MaxTextureUnits;
	ctx->Const.MaxTextureCoordUnits = ctx->Const.MaxTextureUnits;

	driCalculateMaxTextureLevels(rmesa->texture_heaps, rmesa->nr_heaps, &ctx->Const, 4, 11,	/* max 2D texture size is 2048x2048 */
#if ENABLE_HW_3D_TEXTURE
				     8,	/* max 3D texture size is 256^3 */
#else
				     0,	/* 3D textures unsupported */
#endif
				     11,	/* max cube texture size is 2048x2048 */
				     11,	/* max texture rectangle size is 2048x2048 */
				     12, GL_FALSE);

	ctx->Const.MaxTextureMaxAnisotropy = 16.0;

	/* No wide points.
	 */
	ctx->Const.MinPointSize = 1.0;
	ctx->Const.MinPointSizeAA = 1.0;
	ctx->Const.MaxPointSize = 1.0;
	ctx->Const.MaxPointSizeAA = 1.0;

	ctx->Const.MinLineWidth = 1.0;
	ctx->Const.MinLineWidthAA = 1.0;
	ctx->Const.MaxLineWidth = 10.0;
	ctx->Const.MaxLineWidthAA = 10.0;
	ctx->Const.LineWidthGranularity = 0.0625;

	/* Initialize the software rasterizer and helper modules.
	 */
	_swrast_CreateContext(ctx);
	_ac_CreateContext(ctx);
	_tnl_CreateContext(ctx);
	_swsetup_CreateContext(ctx);
	_ae_create_context(ctx);

	/* Install the customized pipeline:
	 */
	_tnl_destroy_pipeline(ctx);
	_tnl_install_pipeline(ctx, r200_pipeline);
	ctx->Driver.FlushVertices = r200FlushVertices;

	/* Try and keep materials and vertices separate:
	 */
	_tnl_isolate_materials(ctx, GL_TRUE);

	/* Configure swrast and TNL to match hardware characteristics:
	 */
	_swrast_allow_pixel_fog(ctx, GL_FALSE);
	_swrast_allow_vertex_fog(ctx, GL_TRUE);
	_tnl_allow_pixel_fog(ctx, GL_FALSE);
	_tnl_allow_vertex_fog(ctx, GL_TRUE);

	for (i = 0; i < R200_MAX_TEXTURE_UNITS; i++) {
		_math_matrix_ctr(&rmesa->TexGenMatrix[i]);
		_math_matrix_set_identity(&rmesa->TexGenMatrix[i]);
	}
	_math_matrix_ctr(&rmesa->tmpmat);
	_math_matrix_set_identity(&rmesa->tmpmat);

	driInitExtensions(ctx, card_extensions, GL_TRUE);
	if (rmesa->radeon.radeonScreen->drmSupportsCubeMaps)
		_mesa_enable_extension(ctx, "GL_ARB_texture_cube_map");
	if (rmesa->radeon.radeonScreen->drmSupportsBlendColor) {
		_mesa_enable_extension(ctx, "GL_EXT_blend_equation_separate");
		_mesa_enable_extension(ctx, "GL_EXT_blend_func_separate");
	}
	if (driQueryOptionb(&rmesa->radeon.optionCache, "arb_vertex_program"))
		_mesa_enable_extension(ctx, "GL_ARB_vertex_program");
	if (driQueryOptionb(&rmesa->radeon.optionCache, "nv_vertex_program"))
		_mesa_enable_extension(ctx, "GL_NV_vertex_program");

#if 0
	r200InitDriverFuncs(ctx);
	r200InitIoctlFuncs(ctx);
	r200InitStateFuncs(ctx);
	r200InitTextureFuncs(ctx);
#endif
	/* plug in a few more device driver functions */
	/* XXX these should really go right after _mesa_init_driver_functions() */
	r200InitPixelFuncs(ctx);
	radeonInitSpanFuncs(ctx);
	r200InitTnlFuncs(ctx);
	r200InitState(rmesa);
	r200InitSwtcl(ctx);

	rmesa->prefer_gart_client_texturing =
	    (getenv("R200_GART_CLIENT_TEXTURES") != 0);

	tcl_mode = driQueryOptioni(&rmesa->radeon.optionCache, "tcl_mode");
	if (IS_FAMILY_R300(&rmesa->radeon) ||
	    driQueryOptionb(&rmesa->radeon.optionCache, "no_rast")) {
		fprintf(stderr, "disabling 3D acceleration\n");
		FALLBACK(&rmesa->radeon, RADEON_FALLBACK_DISABLE, 1);
	}
	if (tcl_mode == DRI_CONF_TCL_SW || getenv("R200_NO_TCL") ||
		   !(rmesa->radeon.radeonScreen->chipset & RADEON_CHIPSET_TCL)) {
		if (rmesa->radeon.radeonScreen->chipset & RADEON_CHIPSET_TCL) {
			rmesa->radeon.radeonScreen->chipset &= ~RADEON_CHIPSET_TCL;
			fprintf(stderr, "Disabling HW TCL support\n");
		}
		TCL_FALLBACK(rmesa->radeon.glCtx, RADEON_TCL_FALLBACK_TCL_DISABLE, 1);
	}
	if (rmesa->radeon.radeonScreen->chipset & RADEON_CHIPSET_TCL) {
		if (tcl_mode >= DRI_CONF_TCL_VTXFMT
		    && !getenv("R200_NO_VTXFMT")) {
			r200VtxfmtInit(ctx, tcl_mode >= DRI_CONF_TCL_CODEGEN);
		}
		_tnl_need_dlist_norm_lengths(ctx, GL_FALSE);
	}
	return GL_TRUE;
}

/* Destroy the device specific context.
 */
/* Destroy the Mesa and driver specific context data.
 */
void r200DestroyContext(__DRIcontextPrivate * driContextPriv)
{
	GET_CURRENT_CONTEXT(ctx);
	r200ContextPtr rmesa = (r200ContextPtr) driContextPriv->driverPrivate;
	radeonContextPtr current = ctx ? RADEON_CONTEXT(ctx) : NULL;

	/* check if we're deleting the currently bound context */
	if (&rmesa->radeon == current) {
		R200_FIREVERTICES(rmesa);
		_mesa_make_current2(NULL, NULL, NULL);
	}

	/* Free r200 context resources */
	assert(rmesa);		/* should never be null */
	if (rmesa) {
		GLboolean release_texture_heaps;

		release_texture_heaps = (rmesa->radeon.glCtx->Shared->RefCount == 1);
		_swsetup_DestroyContext(rmesa->radeon.glCtx);
		_tnl_DestroyContext(rmesa->radeon.glCtx);
		_ac_DestroyContext(rmesa->radeon.glCtx);
		_swrast_DestroyContext(rmesa->radeon.glCtx);

		r200DestroySwtcl(rmesa->radeon.glCtx);
		r200ReleaseArrays(rmesa->radeon.glCtx, ~0);

		if (rmesa->dma.current.buf) {
			r200ReleaseDmaRegion(rmesa, &rmesa->dma.current,
					     __FUNCTION__);
			r200FlushCmdBuf(rmesa, __FUNCTION__);
		}

		if (!(rmesa->radeon.TclFallback & RADEON_TCL_FALLBACK_TCL_DISABLE)) {
			int tcl_mode =
			    driQueryOptioni(&rmesa->radeon.optionCache, "tcl_mode");
			if (tcl_mode >= DRI_CONF_TCL_VTXFMT)
				r200VtxfmtDestroy(rmesa->radeon.glCtx);
		}

		radeonCleanupContext(&rmesa->radeon);

		if (release_texture_heaps) {
			/* This share group is about to go away, free our private
			 * texture object data.
			 */
			int i;

			for (i = 0; i < rmesa->nr_heaps; i++) {
				driDestroyTextureHeap(rmesa->texture_heaps[i]);
				rmesa->texture_heaps[i] = NULL;
			}

			assert(is_empty_list(&rmesa->swapped));
		}

		/* free the option cache */
		driDestroyOptionCache(&rmesa->radeon.optionCache);

		FREE(rmesa);
	}
}
