/*
Copyright (C) The Weather Channel, Inc.  2002.  All Rights Reserved.

The Weather Channel (TM) funded Tungsten Graphics to develop the
initial release of the Radeon 8500 driver under the XFree86 license.
This notice must be preserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE COPYRIGHT OWNER(S) AND/OR ITS SUPPLIERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

**************************************************************************/

/*
 * Authors:
 *   Keith Whitwell <keith@tungstengraphics.com>
 */

#include <sched.h>
#include <errno.h>

#include "glheader.h"
#include "imports.h"
#include "macros.h"
#include "context.h"
#include "swrast/swrast.h"

#include "r200_context.h"
#include "r200_state.h"
#include "radeon_ioctl.h"
#include "r200_ioctl.h"
#include "r200_tcl.h"
#include "r200_sanity.h"
#include "radeon_reg.h"

#include "vblank.h"

#define R200_TIMEOUT             512
#define R200_IDLE_RETRY           16


void r200SaveHwState(r200ContextPtr r200)
{
	struct r200_state_atom *atom;

	foreach(atom, &r200->hw.atomlist)
	    memcpy(atom->savedcmd, atom->cmd, atom->cmd_size * 4);
}

static void r200SwapHwState(r200ContextPtr rmesa)
{
	int *temp;
	struct r200_state_atom *atom;

	foreach(atom, &rmesa->hw.atomlist) {
		temp = atom->cmd;
		atom->cmd = atom->savedcmd;
		atom->savedcmd = temp;
	}
}

/* At this point we were in FlushCmdBufLocked but we had lost our context, so
 * we need to unwire our current cmdbuf and hook a new one in, emit that, then
 * wire the old cmdbuf back in so that FlushCmdBufLocked can continue and the
 * buffer can depend on the state not being lost across lock/unlock.
 */
static void r200BackUpAndEmitLostStateLocked(r200ContextPtr rmesa)
{
	GLuint nr_released_bufs;
	struct r200_store store;
	struct r200_hw_state temp_state;
	static int count = 0;

	rmesa->radeon.lost_context = GL_FALSE;

	nr_released_bufs = rmesa->dma.nr_released_bufs;
	store = rmesa->store;
	rmesa->store.statenr = 0;
	rmesa->store.primnr = 0;
	rmesa->store.cmd_used = 0;
	rmesa->store.elts_start = 0;
	rmesa->hw.all_dirty = GL_TRUE;
	r200SwapHwState(rmesa);
	/* In this case it's okay to EmitState while locked because we won't exhaust
	 * our (empty) cmdbuf.
	 */
	r200EmitState(rmesa);
	r200FlushCmdBufLocked(rmesa, __FUNCTION__);

	r200SwapHwState(rmesa);
	/* We've just cleared out the dirty flags, so we don't remember what
	 * actually needed to be emitted for the next state emit.
	 */
	rmesa->hw.all_dirty = GL_TRUE;
	rmesa->dma.nr_released_bufs = nr_released_bufs;
	rmesa->store = store;
}

int r200FlushCmdBufLocked(r200ContextPtr rmesa, const char *caller)
{
	int ret, i;
	drm_radeon_cmd_buffer_t cmd;

	if (rmesa->radeon.lost_context)
		r200BackUpAndEmitLostStateLocked(rmesa);

	if (RADEON_DEBUG & DEBUG_IOCTL) {
		fprintf(stderr, "%s from %s\n", __FUNCTION__, caller);

		if (0 & RADEON_DEBUG & DEBUG_VERBOSE)
			for (i = 0; i < rmesa->store.cmd_used; i += 4)
				fprintf(stderr, "%d: %x\n", i / 4,
					*(int *)(&rmesa->store.cmd_buf[i]));
	}

	if (RADEON_DEBUG & DEBUG_DMA)
		fprintf(stderr, "%s: Releasing %d buffers\n", __FUNCTION__,
			rmesa->dma.nr_released_bufs);

	if (RADEON_DEBUG & DEBUG_SANITY) {
		if (rmesa->radeon.state.scissor.enabled)
			ret = r200SanityCmdBuffer(rmesa,
						  rmesa->radeon.state.scissor.numClipRects,
						  rmesa->radeon.state.scissor.pClipRects);
		else
			ret = r200SanityCmdBuffer(rmesa,
						  rmesa->radeon.numClipRects,
						  rmesa->radeon.pClipRects);
		if (ret) {
			fprintf(stderr, "drmSanityCommandWrite: %d\n", ret);
			goto out;
		}
	}

	if (RADEON_DEBUG & DEBUG_MEMORY) {
		if (!driValidateTextureHeaps
		    (rmesa->texture_heaps, rmesa->nr_heaps, &rmesa->swapped)) {
			fprintf(stderr,
				"%s: texture memory is inconsistent - expect "
				"mangled textures\n", __FUNCTION__);
		}
	}

	cmd.bufsz = rmesa->store.cmd_used;
	cmd.buf = rmesa->store.cmd_buf;

	if (rmesa->radeon.state.scissor.enabled) {
		cmd.nbox = rmesa->radeon.state.scissor.numClipRects;
		cmd.boxes = (drm_clip_rect_t *) rmesa->radeon.state.scissor.pClipRects;
	} else {
		cmd.nbox = rmesa->radeon.numClipRects;
		cmd.boxes = (drm_clip_rect_t *) rmesa->radeon.pClipRects;
	}

	ret = drmCommandWrite(rmesa->radeon.dri.fd,
			      DRM_RADEON_CMDBUF, &cmd, sizeof(cmd));

	if (ret)
		fprintf(stderr, "drmCommandWrite: %d\n", ret);

	if (RADEON_DEBUG & DEBUG_SYNC) {
		fprintf(stderr, "\nSyncing in %s\n\n", __FUNCTION__);
		radeonWaitForIdleLocked(&rmesa->radeon);
	}

      out:
	rmesa->store.primnr = 0;
	rmesa->store.statenr = 0;
	rmesa->store.cmd_used = 0;
	rmesa->dma.nr_released_bufs = 0;
	rmesa->save_on_next_unlock = 1;

	return ret;
}

/* Note: does not emit any commands to avoid recursion on
 * r200AllocCmdBuf.
 */
void r200FlushCmdBuf(r200ContextPtr rmesa, const char *caller)
{
	int ret;

	LOCK_HARDWARE(&rmesa->radeon);

	ret = r200FlushCmdBufLocked(rmesa, caller);

	UNLOCK_HARDWARE(&rmesa->radeon);

	if (ret) {
		fprintf(stderr, "drmRadeonCmdBuffer: %d (exiting)\n", ret);
		exit(ret);
	}
}

/* =============================================================
 * Hardware vertex buffer handling
 */

void r200RefillCurrentDmaRegion(r200ContextPtr rmesa)
{
	struct r200_dma_buffer *dmabuf;
	int fd = rmesa->radeon.dri.fd;
	int index = 0;
	int size = 0;
	drmDMAReq dma;
	int ret;

	if (RADEON_DEBUG & (DEBUG_IOCTL | DEBUG_DMA))
		fprintf(stderr, "%s\n", __FUNCTION__);

	if (rmesa->dma.flush) {
		rmesa->dma.flush(rmesa);
	}

	if (rmesa->dma.current.buf)
		r200ReleaseDmaRegion(rmesa, &rmesa->dma.current, __FUNCTION__);

	if (rmesa->dma.nr_released_bufs > 4)
		r200FlushCmdBuf(rmesa, __FUNCTION__);

	dma.context = rmesa->radeon.dri.hwContext;
	dma.send_count = 0;
	dma.send_list = NULL;
	dma.send_sizes = NULL;
	dma.flags = 0;
	dma.request_count = 1;
	dma.request_size = RADEON_BUFFER_SIZE;
	dma.request_list = &index;
	dma.request_sizes = &size;
	dma.granted_count = 0;

	LOCK_HARDWARE(&rmesa->radeon);	/* no need to validate */

	while (1) {
		ret = drmDMA(fd, &dma);
		if (ret == 0)
			break;

		if (rmesa->dma.nr_released_bufs) {
			r200FlushCmdBufLocked(rmesa, __FUNCTION__);
		}

		if (rmesa->radeon.do_usleeps) {
			UNLOCK_HARDWARE(&rmesa->radeon);
			DO_USLEEP(1);
			LOCK_HARDWARE(&rmesa->radeon);
		}
	}

	UNLOCK_HARDWARE(&rmesa->radeon);

	if (RADEON_DEBUG & DEBUG_DMA)
		fprintf(stderr, "Allocated buffer %d\n", index);

	dmabuf = CALLOC_STRUCT(r200_dma_buffer);
	dmabuf->buf = &rmesa->radeon.radeonScreen->buffers->list[index];
	dmabuf->refcount = 1;

	rmesa->dma.current.buf = dmabuf;
	rmesa->dma.current.address = dmabuf->buf->address;
	rmesa->dma.current.end = dmabuf->buf->total;
	rmesa->dma.current.start = 0;
	rmesa->dma.current.ptr = 0;
}

void r200ReleaseDmaRegion(r200ContextPtr rmesa,
			  struct r200_dma_region *region, const char *caller)
{
	if (RADEON_DEBUG & DEBUG_IOCTL)
		fprintf(stderr, "%s from %s\n", __FUNCTION__, caller);

	if (!region->buf)
		return;

	if (rmesa->dma.flush)
		rmesa->dma.flush(rmesa);

	if (--region->buf->refcount == 0) {
		drm_radeon_cmd_header_t *cmd;

		if (RADEON_DEBUG & (DEBUG_IOCTL | DEBUG_DMA))
			fprintf(stderr, "%s -- DISCARD BUF %d\n", __FUNCTION__,
				region->buf->buf->idx);

		cmd =
		    (drm_radeon_cmd_header_t *) r200AllocCmdBuf(rmesa,
								sizeof(*cmd),
								__FUNCTION__);
		cmd->dma.cmd_type = RADEON_CMD_DMA_DISCARD;
		cmd->dma.buf_idx = region->buf->buf->idx;
		FREE(region->buf);
		rmesa->dma.nr_released_bufs++;
	}

	region->buf = 0;
	region->start = 0;
}

/* Allocates a region from rmesa->dma.current.  If there isn't enough
 * space in current, grab a new buffer (and discard what was left of current)
 */
void r200AllocDmaRegion(r200ContextPtr rmesa,
			struct r200_dma_region *region,
			int bytes, int alignment)
{
	if (RADEON_DEBUG & DEBUG_IOCTL)
		fprintf(stderr, "%s %d\n", __FUNCTION__, bytes);

	if (rmesa->dma.flush)
		rmesa->dma.flush(rmesa);

	if (region->buf)
		r200ReleaseDmaRegion(rmesa, region, __FUNCTION__);

	alignment--;
	rmesa->dma.current.start = rmesa->dma.current.ptr =
	    (rmesa->dma.current.ptr + alignment) & ~alignment;

	if (rmesa->dma.current.ptr + bytes > rmesa->dma.current.end)
		r200RefillCurrentDmaRegion(rmesa);

	region->start = rmesa->dma.current.start;
	region->ptr = rmesa->dma.current.start;
	region->end = rmesa->dma.current.start + bytes;
	region->address = rmesa->dma.current.address;
	region->buf = rmesa->dma.current.buf;
	region->buf->refcount++;

	rmesa->dma.current.ptr += bytes;	/* bug - if alignment > 7 */
	rmesa->dma.current.start =
	    rmesa->dma.current.ptr = (rmesa->dma.current.ptr + 0x7) & ~0x7;

	assert(rmesa->dma.current.ptr <= rmesa->dma.current.end);
}

void r200AllocDmaRegionVerts(r200ContextPtr rmesa,
			     struct r200_dma_region *region,
			     int numverts, int vertsize, int alignment)
{
	r200AllocDmaRegion(rmesa, region, vertsize * numverts, alignment);
}


/* ================================================================
 * Buffer clear
 */
static void r200Clear(GLcontext * ctx, GLbitfield mask, GLboolean all,
		      GLint cx, GLint cy, GLint cw, GLint ch)
{
	r200ContextPtr rmesa = R200_CONTEXT(ctx);
	__DRIdrawablePrivate *dPriv = rmesa->radeon.dri.drawable;
	GLuint flags = 0;
	GLint ret, i;

	if (RADEON_DEBUG & DEBUG_IOCTL) {
		fprintf(stderr, "%s:  all=%d cx=%d cy=%d cw=%d ch=%d\n",
			__FUNCTION__, all, cx, cy, cw, ch);
	}

	{
		LOCK_HARDWARE(&rmesa->radeon);
		UNLOCK_HARDWARE(&rmesa->radeon);
		if (dPriv->numClipRects == 0)
			return;
	}

	R200_FIREVERTICES(rmesa);

	if (mask & DD_FRONT_LEFT_BIT) {
		flags |= RADEON_FRONT;
		mask &= ~DD_FRONT_LEFT_BIT;
	}

	if (mask & DD_BACK_LEFT_BIT) {
		flags |= RADEON_BACK;
		mask &= ~DD_BACK_LEFT_BIT;
	}

	if (mask & DD_DEPTH_BIT) {
		if (ctx->Depth.Mask)
			flags |= RADEON_DEPTH;	/* FIXME: ??? */
		mask &= ~DD_DEPTH_BIT;
	}

	if ((mask & DD_STENCIL_BIT) && rmesa->state.stencil.hwBuffer) {
		flags |= RADEON_STENCIL;
		mask &= ~DD_STENCIL_BIT;
	}

	if (mask) {
		if (RADEON_DEBUG & DEBUG_FALLBACKS)
			fprintf(stderr, "%s: swrast clear, mask: %x\n",
				__FUNCTION__, mask);
		_swrast_Clear(ctx, mask, all, cx, cy, cw, ch);
	}

	if (!flags)
		return;

	/* Flip top to bottom */
	cx += dPriv->x;
	cy = dPriv->y + dPriv->h - cy - ch;

	LOCK_HARDWARE(&rmesa->radeon);

	/* Throttle the number of clear ioctls we do.
	 */
	while (1) {
		drm_radeon_getparam_t gp;
		int ret;
		int clear;

		gp.param = RADEON_PARAM_LAST_CLEAR;
		gp.value = (int *)&clear;
		ret = drmCommandWriteRead(rmesa->radeon.dri.fd,
					  DRM_RADEON_GETPARAM, &gp, sizeof(gp));

		if (ret) {
			fprintf(stderr, "%s: drmRadeonGetParam: %d\n",
				__FUNCTION__, ret);
			exit(1);
		}

		/* Clear throttling needs more thought.
		 */
		if (rmesa->radeon.sarea->last_clear - clear <= 25) {
			break;
		}

		if (rmesa->radeon.do_usleeps) {
			UNLOCK_HARDWARE(&rmesa->radeon);
			DO_USLEEP(1);
			LOCK_HARDWARE(&rmesa->radeon);
		}
	}

	/* Send current state to the hardware */
	r200FlushCmdBufLocked(rmesa, __FUNCTION__);

	for (i = 0; i < dPriv->numClipRects;) {
		GLint nr =
		    MIN2(i + RADEON_NR_SAREA_CLIPRECTS, dPriv->numClipRects);
		drm_clip_rect_t *box = dPriv->pClipRects;
		drm_clip_rect_t *b = rmesa->radeon.sarea->boxes;
		drm_radeon_clear_t clear;
		drm_radeon_clear_rect_t depth_boxes[RADEON_NR_SAREA_CLIPRECTS];
		GLint n = 0;

		if (!all) {
			for (; i < nr; i++) {
				GLint x = box[i].x1;
				GLint y = box[i].y1;
				GLint w = box[i].x2 - x;
				GLint h = box[i].y2 - y;

				if (x < cx)
					w -= cx - x, x = cx;
				if (y < cy)
					h -= cy - y, y = cy;
				if (x + w > cx + cw)
					w = cx + cw - x;
				if (y + h > cy + ch)
					h = cy + ch - y;
				if (w <= 0)
					continue;
				if (h <= 0)
					continue;

				b->x1 = x;
				b->y1 = y;
				b->x2 = x + w;
				b->y2 = y + h;
				b++;
				n++;
			}
		} else {
			for (; i < nr; i++) {
				*b++ = box[i];
				n++;
			}
		}

		rmesa->radeon.sarea->nbox = n;

		clear.flags = flags;
		clear.clear_color = rmesa->radeon.state.color.clear;
		clear.clear_depth = 0;	/* not used */
		clear.color_mask = rmesa->hw.msk.cmd[MSK_RB3D_PLANEMASK];
		clear.depth_mask = rmesa->state.stencil.clear;
		clear.depth_boxes = depth_boxes;

		n--;
		b = rmesa->radeon.sarea->boxes;
		for (; n >= 0; n--) {
			depth_boxes[n].f[CLEAR_X1] = (float)b[n].x1;
			depth_boxes[n].f[CLEAR_Y1] = (float)b[n].y1;
			depth_boxes[n].f[CLEAR_X2] = (float)b[n].x2;
			depth_boxes[n].f[CLEAR_Y2] = (float)b[n].y2;
			depth_boxes[n].f[CLEAR_DEPTH] = ctx->Depth.Clear;
		}

		ret = drmCommandWrite(rmesa->radeon.dri.fd, DRM_RADEON_CLEAR,
				      &clear, sizeof(clear));

		if (ret) {
			UNLOCK_HARDWARE(&rmesa->radeon);
			fprintf(stderr, "DRM_RADEON_CLEAR: return = %d\n", ret);
			exit(1);
		}
	}

	UNLOCK_HARDWARE(&rmesa->radeon);
	rmesa->hw.all_dirty = GL_TRUE;
}

void r200Flush(GLcontext * ctx)
{
	r200ContextPtr r200 = R200_CONTEXT(ctx);

	if (RADEON_DEBUG & DEBUG_IOCTL)
		fprintf(stderr, "%s\n", __FUNCTION__);

	if (r200->dma.flush)
		r200->dma.flush(r200);

	r200EmitState(r200);

	if (r200->store.cmd_used) {
		r200FlushCmdBuf(r200, __FUNCTION__);
	}
}


/* This version of AllocateMemoryMESA allocates only GART memory, and
 * only does so after the point at which the driver has been
 * initialized.
 *
 * Theoretically a valid context isn't required.  However, in this
 * implementation, it is, as I'm using the hardware lock to protect
 * the kernel data structures, and the current context to get the
 * device fd.
 */
void *r200AllocateMemoryMESA(__DRInativeDisplay * dpy, int scrn, GLsizei size,
			     GLfloat readfreq, GLfloat writefreq,
			     GLfloat priority)
{
	GET_CURRENT_CONTEXT(ctx);
	r200ContextPtr rmesa;
	int region_offset;
	drm_radeon_mem_alloc_t alloc;
	int ret;

	if (RADEON_DEBUG & DEBUG_IOCTL)
		fprintf(stderr, "%s sz %d %f/%f/%f\n", __FUNCTION__, size,
			readfreq, writefreq, priority);

	if (!ctx || !(rmesa = R200_CONTEXT(ctx))
	    || !rmesa->radeon.radeonScreen->gartTextures.map)
		return NULL;

	if (getenv("R200_NO_ALLOC"))
		return NULL;

	if (rmesa->radeon.dri.drmMinor < 6)
		return NULL;

	alloc.region = RADEON_MEM_REGION_GART;
	alloc.alignment = 0;
	alloc.size = size;
	alloc.region_offset = &region_offset;

	ret = drmCommandWriteRead(rmesa->radeon.radeonScreen->driScreen->fd,
				  DRM_RADEON_ALLOC, &alloc, sizeof(alloc));

	if (ret) {
		fprintf(stderr, "%s: DRM_RADEON_ALLOC ret %d\n", __FUNCTION__,
			ret);
		return NULL;
	}

	{
		char *region_start =
		    (char *)rmesa->radeon.radeonScreen->gartTextures.map;
		return (void *)(region_start + region_offset);
	}
}

/* Called via glXFreeMemoryMESA() */
void r200FreeMemoryMESA(__DRInativeDisplay * dpy, int scrn, GLvoid * pointer)
{
	GET_CURRENT_CONTEXT(ctx);
	r200ContextPtr rmesa;
	int region_offset;
	drm_radeon_mem_free_t memfree;
	int ret;

	if (RADEON_DEBUG & DEBUG_IOCTL)
		fprintf(stderr, "%s %p\n", __FUNCTION__, pointer);

	if (!ctx || !(rmesa = R200_CONTEXT(ctx))
	    || !rmesa->radeon.radeonScreen->gartTextures.map) {
		fprintf(stderr, "%s: no context\n", __FUNCTION__);
		return;
	}

	if (rmesa->radeon.dri.drmMinor < 6)
		return;

	region_offset =
	    (char *)pointer - (char *)rmesa->radeon.radeonScreen->gartTextures.map;

	if (region_offset < 0 ||
	    region_offset > rmesa->radeon.radeonScreen->gartTextures.size) {
		fprintf(stderr, "offset %d outside range 0..%d\n",
			region_offset, rmesa->radeon.radeonScreen->gartTextures.size);
		return;
	}

	memfree.region = RADEON_MEM_REGION_GART;
	memfree.region_offset = region_offset;

	ret = drmCommandWrite(rmesa->radeon.radeonScreen->driScreen->fd,
			      DRM_RADEON_FREE, &memfree, sizeof(memfree));

	if (ret)
		fprintf(stderr, "%s: DRM_RADEON_FREE ret %d\n", __FUNCTION__,
			ret);
}

/* Called via glXGetMemoryOffsetMESA() */
GLuint r200GetMemoryOffsetMESA(__DRInativeDisplay * dpy, int scrn,
			       const GLvoid * pointer)
{
	GET_CURRENT_CONTEXT(ctx);
	r200ContextPtr rmesa;
	GLuint card_offset;

	if (!ctx || !(rmesa = R200_CONTEXT(ctx))) {
		fprintf(stderr, "%s: no context\n", __FUNCTION__);
		return ~0;
	}

	if (!r200IsGartMemory(rmesa, pointer, 0))
		return ~0;

	if (rmesa->radeon.dri.drmMinor < 6)
		return ~0;

	card_offset = r200GartOffsetFromVirtual(rmesa, pointer);

	return card_offset - rmesa->radeon.radeonScreen->gart_base;
}

GLboolean r200IsGartMemory(r200ContextPtr rmesa, const GLvoid * pointer,
			   GLint size)
{
	int offset =
	    (char *)pointer - (char *)rmesa->radeon.radeonScreen->gartTextures.map;
	int valid = (size >= 0 && offset >= 0
		     && offset + size < rmesa->radeon.radeonScreen->gartTextures.size);

	if (RADEON_DEBUG & DEBUG_IOCTL)
		fprintf(stderr, "r200IsGartMemory( %p ) : %d\n", pointer,
			valid);

	return valid;
}

GLuint r200GartOffsetFromVirtual(r200ContextPtr rmesa, const GLvoid * pointer)
{
	int offset =
	    (char *)pointer - (char *)rmesa->radeon.radeonScreen->gartTextures.map;

	if (offset < 0 || offset > rmesa->radeon.radeonScreen->gartTextures.size)
		return ~0;
	else
		return rmesa->radeon.radeonScreen->gart_texture_offset + offset;
}

void r200InitIoctlFuncs(struct dd_function_table *functions)
{
	functions->Clear = r200Clear;
	functions->Finish = radeonFinish;
	functions->Flush = r200Flush;
}

